#include "ramtron.h"

void i2c_init (void)
{
	// Configure port C pins 3 and 4 to "analogue" pins.
	trisc.TRISC3 = 1;
	trisc.TRISC4 = 1;

	// Disable slew rate control and SMBus pins.
	sspstat.SMP = 1;
	sspstat.CKE = 0;

	// Enable MSSP, configure as a i2c master
	sspcon1 = 0x28;

	// Baud rate configuration
	// ((40 Mhz / 100kHz) / 4) - 1 = 99 = 0x63;
	sspadd = 0x63;

	// Flush control bits 
	sspcon2 = 0x00;
}

void fm3164_init (void)
{
	// Enable oscillator
	fm3164_write_register(FM3164_REGISTER_CONTROL, 0x00);

	// Clear the 'low voltage' and other watchdog bits.
	fm3164_write_register(FM3164_REGISTER_WATCHDOG, 0x00);

	// Halt the clock (set write bit)
	fm3164_write_register(FM3164_REGISTER_FLAGS, 0x02);
/*
	fm3164_write_register(FM3164_REGISTER_YEAR, 0x08);
	fm3164_write_register(FM3164_REGISTER_MONTH, 0x11);
	fm3164_write_register(FM3164_REGISTER_DAY, 0x15);
	fm3164_write_register(FM3164_REGISTER_HOURS, 0x14);
	fm3164_write_register(FM3164_REGISTER_MINUTES, 0x36);
	fm3164_write_register(FM3164_REGISTER_SECONDS, 0x33);


	fm3164_write_register(FM3164_REGISTER_YEAR, 0x00);
	fm3164_write_register(FM3164_REGISTER_MONTH, 0x00);
	fm3164_write_register(FM3164_REGISTER_DAY, 0x00);
	fm3164_write_register(FM3164_REGISTER_HOURS, 0x00);
	fm3164_write_register(FM3164_REGISTER_MINUTES, 0x00);
	fm3164_write_register(FM3164_REGISTER_SECONDS, 0x00);
*/
	// Re-enable the clock (clear write)
	fm3164_write_register(FM3164_REGISTER_FLAGS, 0x00);
}

i2c_status_t i2c_write (unsigned char data)
{
	sspcon1.WCOL = 0;

	// Wait for MSSP/I2C transmitter to become idle.
	while (sspstat.R_W || sspcon2.SEN || sspcon2.RSEN || sspcon2.RCEN || sspcon2.ACKEN);

	sspbuf = data;
	if (sspcon1.WCOL) { return I2C_WRITE_COLLISION; }
	//while (sspstat.BF);

	// Wait for transmission to finish.
	while (sspstat.R_W || sspcon2.SEN || sspcon2.RSEN || sspcon2.RCEN || sspcon2.ACKEN || !pir1.SSPIF);
	
	I2C_IDLE_WAIT();

	if (sspcon2.ACKSTAT) { return I2C_NO_ACK; }

	return I2C_NO_ERROR;
}

i2c_status_t fm3164_read_memory (unsigned int address, unsigned char *dst, unsigned char length)
{
	i2c_status_t status;

	for ( ; length > 0; --length) {
		if ((status = fm3164_read_memory_byte(address++, dst++)) != I2C_NO_ERROR) {
			return status;
		}
	}

	return I2C_NO_ERROR;
}

i2c_status_t fm3164_read_memory_byte (unsigned int address, unsigned char *dst)
{
	i2c_status_t status;
	unsigned char tmp;

	#ifdef FM3164_DEBUG
	unsigned char buffer[32];
	#endif

	// Issue initial start condition.
	I2C_IDLE_WAIT();
	sspcon2.SEN = 1;
	while (sspcon2.SEN);

	if (address < FM3164_MEMORY_SIZE) {
		// Write control byte (set memory address)
		if ((status = i2c_write(0xA0)) != I2C_NO_ERROR) {
			return status;
		}

		// Write first address byte (MSB)
		tmp = address >> 8;
		if ((status = i2c_write(tmp)) != I2C_NO_ERROR) {
			return status;
		}

		// Write the second address byte (LSB)
		tmp = address;
		if ((status = i2c_write(tmp)) != I2C_NO_ERROR) {
			return status;
		}

		// Issue restart condition for readback operation.
		//I2C_IDLE_WAIT();
		sspcon2.RSEN = 1;
		while (sspcon2.RSEN);
	}

	// Write control byte.
	if ((status = i2c_write(0xA1)) != I2C_NO_ERROR) {
		return status;
	}

	i2c_get_bytes(dst, 1);

	// Issue stop condition
	sspcon2.PEN = 1;
	while(sspcon2.PEN);

	#ifdef FM3164_DEBUG
	sprintf(buffer, "i2c/r: address: 0x%X ", address);
	uart_send_string(buffer);
	sprintf(buffer, "data: 0x%X\r\n", *dst);
	uart_send_string(buffer);
	#endif

	return status;
}

i2c_status_t fm3164_write_memory (unsigned int address,
	unsigned char *data, unsigned char length)
{
	i2c_status_t status;

	for (; length > 0; --length) {
		if ((status = fm3164_write_byte_memory(address++, *data++)) != I2C_NO_ERROR) {
			return status;
		}
	}
}

/*
NOTE: Fm3164 requires the MSB byte first for internal memory addressing.
*/
i2c_status_t fm3164_write_byte_memory (unsigned int address, unsigned char data)
{
	i2c_status_t status;
	unsigned char temp;

	#ifdef FM3164_DEBUG
	unsigned char buffer[32];

	// Debug output, change '0' above to '1' to enable.
	sprintf(buffer, "i2c/w: address: 0x%X ", address);
	uart_send_string(buffer);
	sprintf(buffer, "data: 0x%X\r\n", data);
	uart_send_string(buffer);
	#endif

	// Issue initial start condition.
	I2C_IDLE_WAIT();
	sspcon2.SEN = 1;
	while (sspcon2.SEN);

	// Write control byte.
	if ((status = i2c_write(0xA0)) != I2C_NO_ERROR) {
		return status;
	}
	I2C_IDLE_WAIT();

	// Write first address byte MSB
	temp = address >> 8;
	if ((status = i2c_write(temp)) != I2C_NO_ERROR) {
		return status;
	}
	// Write the second address byte LSB
	temp = address;
	if ((status = i2c_write(temp)) != I2C_NO_ERROR) {
		return status;
	}
	
	if ((status = i2c_write(data)) != I2C_NO_ERROR) { return status; }

	// Issue stop condition
	sspcon2.PEN = 1;
	while(sspcon2.PEN);

	return status;
}

void i2c_get_bytes (unsigned char *data, unsigned char length)
{
	for ( ; length > 0; --length) {

		I2C_IDLE_WAIT();

		// Read the response.
		sspcon2.RCEN = 1;
		while (!sspstat.BF);
		sspcon2.RCEN = 0;

		*data++ = sspbuf;
		// Send NACK if this is the last byte.
		if (length == 1) { sspcon2.ACKDT = 1; }

		sspcon2.ACKEN = 1;
		while (sspcon2.ACKEN);
	}
	
	sspcon2.ACKDT = 0;
}

void fm3164_set_date (fm3164_date_t *date)
{
	unsigned char value;

	#ifdef FM3164_DEBUG
	unsigned char buffer[12];
	#endif

	// Set the write bit to one in order to halt the clock.
	fm3164_read_register(FM3164_REGISTER_FLAGS, &value);
	value |= 0x02;
	fm3164_write_register(FM3164_REGISTER_FLAGS, value);

	#ifdef FM3164_DEBUG
	uart_send_string("write bit in flags: ");
	sprintf(buffer, "0x%X\r\n", value);
	uart_send_string(buffer);
	#endif

	// Write in the year to the FM3104.
	value = bin_to_bcd(date->year);
	fm3164_write_register(FM3164_REGISTER_YEAR, value);

	#ifdef FM3164_DEBUG
	uart_send_string("year: ");
	sprintf(buffer, "0x%X\r\n", value);
	uart_send_string(buffer);
	#endif

	// Write in the months.
	value = bin_to_bcd(date->month);
	fm3164_write_register(FM3164_REGISTER_MONTH, value);

	#ifdef FM3164_DEBUG
	uart_send_string("month: ");
	sprintf(buffer, "0x%X\r\n", value);
	uart_send_string(buffer);
	#endif

	// Write in the days.
	value = bin_to_bcd(date->day);
	fm3164_write_register(FM3164_REGISTER_DAY, value);

	#ifdef FM3164_DEBUG
	uart_send_string("day: ");
	sprintf(buffer, "0x%X\r\n", value);
	uart_send_string(buffer);
	#endif

	// Write in the hours.
	value = bin_to_bcd(date->hours);
	fm3164_write_register(FM3164_REGISTER_HOURS, value);

	#ifdef FM3164_DEBUG
	uart_send_string("hours: ");
	sprintf(buffer, "0x%X\r\n", value);
	uart_send_string(buffer);
	#endif

	// Write in the minutes
	value = bin_to_bcd(date->minutes);
	fm3164_write_register(FM3164_REGISTER_MINUTES, value);

	#ifdef FM3164_DEBUG
	uart_send_string("minutes: ");
	sprintf(buffer, "0x%X\r\n", value);
	uart_send_string(buffer);
	#endif

	// Write in the seconds
	value = bin_to_bcd(date->seconds);
	fm3164_write_register(FM3164_REGISTER_SECONDS, value);

	#ifdef FM3164_DEBUG
	uart_send_string("seconds: ");
	sprintf(buffer, "0x%X\r\n", value);
	uart_send_string(buffer);
	#endif

	// Set the write bit to zero in order to start the clock again..
	fm3164_read_register(FM3164_REGISTER_FLAGS, &value);

	value &= ~0x02;

	#ifdef FM3164_DEBUG
	uart_send_string("flags value after modifications: ");
	sprintf(buffer, "0x%X\r\n", value);
	uart_send_string(buffer);
	#endif

	fm3164_write_register(FM3164_REGISTER_FLAGS, value);

}

void fm3164_get_date (fm3164_date_t *date)
{
	unsigned char value;

	// Set the read bit to one in order to latch the date counter.
	fm3164_write_register(FM3164_REGISTER_FLAGS, 0x01);

	fm3164_read_register(FM3164_REGISTER_YEAR, &value);
	date->year = bcd_to_bin(value);

	fm3164_read_register(FM3164_REGISTER_MONTH, &value);
	date->month = bcd_to_bin(value);

	fm3164_read_register(FM3164_REGISTER_DAY, &value);
	date->day = bcd_to_bin(value);

	fm3164_read_register(FM3164_REGISTER_HOURS, &value);
	date->hours = bcd_to_bin(value);

	fm3164_read_register(FM3164_REGISTER_MINUTES, &value);
	date->minutes = bcd_to_bin(value);

	fm3164_read_register(FM3164_REGISTER_SECONDS, &value);
	date->seconds = bcd_to_bin(value);

	// Set the read bit to zero now that we are finished reading.
	fm3164_write_register(FM3164_REGISTER_FLAGS, 0x00);
}

i2c_status_t fm3164_write_register (
	unsigned char address, unsigned char data)
{
	i2c_status_t status;

	// Issue initial start condition.
	I2C_IDLE_WAIT();
	sspcon2.SEN = 1;
	while (sspcon2.SEN);

	// Write control byte (0xD0 = write access to companion).
	if ((status = i2c_write(0xD0)) != I2C_NO_ERROR) {
		return status;
	}
	I2C_IDLE_WAIT();

	// Write the address byte.
	if ((status = i2c_write(address)) != I2C_NO_ERROR) {
		return status;
	}

	// Write the data value byte.
	if ((status = i2c_write(data)) != I2C_NO_ERROR) {
		return status;
	}

	// Issue stop condition
	sspcon2.PEN = 1;
	while(sspcon2.PEN);

	return status;
}

i2c_status_t fm3164_read_register (
	unsigned char address, unsigned char *data)
{
	i2c_status_t status;

	// Issue initial start condition.
	I2C_IDLE_WAIT();
	sspcon2.SEN = 1;
	while (sspcon2.SEN);

	// Write control byte (write access to companion device).
	if ((status = i2c_write(0xD0)) != I2C_NO_ERROR) {
		return status;
	}

	// Write the address byte (which register to access).
	if ((status = i2c_write(address)) != I2C_NO_ERROR) {
		return status;
	}

	// Issue restart condition for readback operation.
	//I2C_IDLE_WAIT();
	sspcon2.RSEN = 1;
	while (sspcon2.RSEN);

	// Write a new control byte (read access to companion device).
	if ((status = i2c_write(0xD1)) != I2C_NO_ERROR) {
		return status;
	}

	i2c_get_bytes(data, 1);

	// Issue stop condition
	sspcon2.PEN = 1;
	while(sspcon2.PEN);

	return status;
}
